from pyrelational.datasets.classification import (
    BreastCancerDataset,
    Checkerboard2x2Dataset,
    Checkerboard4x4Dataset,
    CreditCardDataset,
    DigitDataset,
    FashionMNIST,
    GaussianCloudsDataset,
    StriatumDataset,
    SynthClass1,
    SynthClass2,
    SynthClass3,
    UCIGlass,
    UCIParkinsons,
    UCISeeds,
)
from pyrelational.datasets.regression import (
    DiabetesDataset,
    SynthReg1,
    SynthReg2,
    UCIAirfoil,
    UCIConcrete,
    UCIEnergy,
    UCIPower,
    UCIWine,
    UCIYacht,
)
from pyrelational.datasets.uci_datasets import UCIDatasets
